/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/**
 * @file
 * qp Statemachine, works as interface between library calls and ds485
 * Protocol stack. Handles the message transfer.
 */
#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>

#include "digitalSTROM/ds485.h"
#include "ds485-qp.h"
#include "digitalSTROM/ds485-const.h"
#include "ds485-signals.h"
#include "unix/ds485-unix.h"
#include "ds485-proto.h"

Q_DEFINE_THIS_MODULE(ds485-proto)

extern Ds485LibInst_t  devInst;

/**
 * Transfer statistic
 */
unsigned int busRxCount = 0;
unsigned int busTxCount = 0;


/**
 * Protocol handler
 */

/* Active object class -----------------------------------------------------*/
typedef struct DsmApiTag {
    QActive super;
} DsmApi;

/* Local objects -----------------------------------------------------------*/
static DsmApi l_dsmApi;   /* the single instance of the dSM API active object */

static QEvent const * l_ds485TxResponseQueueSto[1];
QEQueue l_ds485TxResponseQueue;

static QEvent const * l_ds485RxMessageQueueSto[16];
QEQueue l_ds485RxMessageQueue;

static QState DsmApi_initial(DsmApi *me, QEvent const *e);
static QState DsmApi_top(DsmApi *me, QEvent const *e);
static QState DsmApi_idle(DsmApi *me, QEvent const *e);

void postTxResponse( int error )
{
    DS485ResponseEvt * msg = Q_NEW( DS485ResponseEvt, DS485_TX_RESPONSE_SIG );
    msg->error  = error;
    QActive_postFIFO( AO_DsmApi, (QEvent *)msg );
}

int waitTxResponse()
{
    sem_wait( &devInst.semTx );

    DS485ResponseEvt const * msg;
    msg = (DS485ResponseEvt const *)QEQueue_get(&l_ds485TxResponseQueue);
    if ( NULL==msg ){
        return DS485_INTERNAL_ERROR_EMPTY_TX_QUEUE;
    }
    int error   = msg->error;
    QF_gc((QEvent const *) msg);

    return  error;
}

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_DsmApi = (QActive *)&l_dsmApi;      /* "opaque" AO pointer */

/*..........................................................................*/
void DsmApi_ctor(void)
{
    QEQueue_init(&l_ds485RxMessageQueue, l_ds485RxMessageQueueSto, Q_DIM(l_ds485RxMessageQueueSto));
    QEQueue_init(&l_ds485TxResponseQueue, l_ds485TxResponseQueueSto, Q_DIM(l_ds485TxResponseQueueSto));
    /* set initial state */
    DsmApi *me = &l_dsmApi;
    QActive_ctor(&me->super, (QStateHandler)&DsmApi_initial);

}

void DsmApi_printQueueMargins(void)
{
    fprintf(stderr, "\tds485RxMessageQueue: %u/%lu\n", QEQueue_getNFree(&l_ds485RxMessageQueue), Q_DIM(l_ds485RxMessageQueueSto));
    fprintf(stderr, "\tds485TxResponseQueue: %u/%lu\n", QEQueue_getNFree(&l_ds485TxResponseQueue), Q_DIM(l_ds485TxResponseQueueSto));
}

/*..........................................................................*/
void DsmApi_AddDsidToDs485Message( DS485MessageEvt *msg, dsid_t *dsid )
{
    uint16_t    *data;
    uint8_t     idx;

    data    = (uint16_t*)&dsid;

    for( idx = 0; idx < 6; ++idx ) {
        DS485_SetParameter( msg, *data, idx+1 );
        data++;
    }
}

/*..........................................................................*/
static QState DsmApi_initial(DsmApi *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */
    return Q_TRAN(&DsmApi_idle);
}

/*..........................................................................*/
static QState DsmApi_top(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        /* subscribe events */
        //QActive_subscribe((QActive *)me, DATAMODEL_ZONE_ID_SIG);
        QActive_subscribe((QActive *)me, DS485_ADDRESS_SIG);
        return Q_HANDLED();
    case DS485_ADDRESS_SIG:
        return Q_HANDLED();
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DsmApi_idle(DsmApi *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TX_RESPONSE_SIG:
    {
        DS485ResponseEvt const * msg = (DS485ResponseEvt const *)e;

        if( devInst.sender != NULL  ) devInst.sender( &devInst, msg->error );

        if( devInst.blockingMode ){
            QEQueue_postFIFO(&l_ds485TxResponseQueue, (QEvent const *)msg);
            sem_post( &devInst.semTx );
        }
        return Q_HANDLED();
    }
    case DS485_RX_DATA_SIG:
    {
        DS485LongMessageEvt const * rxMsgEvt = (DS485LongMessageEvt const *)e;
        DS485PacketEvt  *txmsg   = Q_NEW( DS485PacketEvt, -1 );
        if( DS485_IsRequest(rxMsgEvt) ){
            txmsg->packet.containerType   = DS485_CONTAINER_REQUEST;
        } else if (DS485_IsResponse(rxMsgEvt)) {
            txmsg->packet.containerType   = DS485_CONTAINER_RESPONSE;
        } else {
            txmsg->packet.containerType = DS485_CONTAINER_EVENT;
        }

        if( DS485_IsBroadcast(rxMsgEvt) ){
            SetBroadcastDsuid(txmsg->packet.destinationId);
        } else {
            txmsg->packet.destinationId   = rxMsgEvt->destinationAddr;
        }
        txmsg->packet.sourceId   = rxMsgEvt->sourceAddr;

        txmsg->packet.length = DS485_GetPayloadLength( rxMsgEvt ) - 1;
        memcpy( txmsg->packet.raw_data, &rxMsgEvt->message[DS485_DATA_OFFSET], txmsg->packet.length + 1 );

        if( devInst.received != NULL ) devInst.received( &devInst, &txmsg->packet );

        if( QEQueue_getNFree(&l_ds485RxMessageQueue) == 0 ){
            sem_wait( &devInst.semRx );                // remove oldest message in the queue
            QEQueue_get(&l_ds485RxMessageQueue);
        }
        QEQueue_postFIFO(&l_ds485RxMessageQueue, (QEvent const *)txmsg);
        sem_post( &devInst.semRx );

        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DsmApi_top);
}

