/*
 This file is part of the digitalSTROM ds485-core library.

 Copyright (C) 2010 digitalSTROM.org, Schlieren, Switzerland

 digitalSTROM ds485-core library is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation version 2
 of the License.

 digitalSTROM ds485-core library is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with digitalSTROM ds485-core library; if not, write to the Free Software
 Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

/**
 * @file
 * dS485 Protocol Stack, more or less the same as in the dSM
 *
 * last sync at 14.09.2010 with cb8b1af6ca5b119730cb03402b969ebc0ef6e207
 */

#ifdef DAEMON_IMPL
    #include "config.h"
    #include "digitalSTROM/ds485.h"
    #include "digitalSTROM/ds485-const.h"
    #include "ds485-qp.h"
    #include "ds485-signals.h"
    #include "unix/bsp.h"
    #include "unix/callqueue.h"
    #include "unix/logger.h"
    #include "ds485-proto.h"
    #include "digitalSTROM/ds485-core.h"
#else
    #include "ds485.h"
    #include "dsmapi.h"
    #include "dsm.h"
    #include "datamodel.h"
    #include "usart3_isr.h"
    #include "bsp.h"
    #include "stm32f10x.h"
#endif

#if defined(__linux__)
# define BSP_MS_PER_TICKS 10
#elif defined(__APPLE__)
# define BSP_MS_PER_TICKS 10
#elif defined(__CYGWIN__)
# define BSP_MS_PER_TICKS 10
#endif

#include "mmi.h"
#include "crc16_ccitt.h"
#include "bitset8.h"

#include <string.h>
#include <stdlib.h>

Q_DEFINE_THIS_MODULE(ds485)

enum DS485Timeouts {
    /* transmit message timeout ticks */
    TX_MSG_TIMEOUT_TICKS                                = (40/BSP_MS_PER_TICKS),
    /* slave startup timeout ticks */
    SLAVE_STARTUP_TIMEOUT_TICKS                         = (60000/BSP_MS_PER_TICKS),
    /* startup request timeout ticks */
    STARTUP_REQUEST_TIMEOUT_TICKS                       = (60/BSP_MS_PER_TICKS),
    /* join request timeout ticks */
    JOIN_REQUEST_TIMEOUT_TICKS                          = (60/BSP_MS_PER_TICKS),
    /* set address response timeout ticks */
    SET_ADDRESS_RESPONSE_TIMEOUT_TICKS                  = (120/BSP_MS_PER_TICKS),
    /* token min hold timeout ticks */
    TOKEN_MIN_HOLD_TIMEOUT_TICKS                        = (10/BSP_MS_PER_TICKS),
    /* token max hold timeout ticks */
    TOKEN_MAX_HOLD_TIMEOUT_TICKS                        = (50/BSP_MS_PER_TICKS),
    /* token long hold timeout ticks */
    TOKEN_LONG_HOLD_TIMEOUT_TICKS                       = (500/BSP_MS_PER_TICKS),
    /* find station timeout ticks */
    FIND_STATION_TIMEOUT_TICKS                          = ((2000/BSP_MS_PER_TICKS)+JOIN_REQUEST_TIMEOUT_TICKS),
    /* find station timeout ticks for single-master */
    FIND_STATION_SECOND_DEVICE_TIMEOUT_TICKS            = (200/BSP_MS_PER_TICKS),
    /* token hand over timeout ticks */
    TOKEN_HAND_OVER_TIMEOUT_TICKS                       = (100/BSP_MS_PER_TICKS),
    /* token missing timeout ticks in master mode */
    TOKEN_MISSING_MASTER_TIMEOUT_TICKS                  = (10000/BSP_MS_PER_TICKS),
    /* token missing timeout ticks in slave mode */
    TOKEN_MISSING_SLAVE_TIMEOUT_TICKS                   = (30000/BSP_MS_PER_TICKS),
    DSMAPI_RESPONSE_TIMEOUT_TICKS                       = (200/BSP_MS_PER_TICKS),
    GET_ADDRESS_REQUEST_TIMEOUT_TICKS                   = (500/BSP_MS_PER_TICKS),
    WAIT_FOR_STARTUP_RESPONSE_TIMEOUT_TICKS             = (250/BSP_MS_PER_TICKS),
    WAIT_FOR_JOIN_RESPONSE_TIMEOUT_TICKS                = (250/BSP_MS_PER_TICKS),
    WAIT_FOR_SET_ADDRESS_REQUEST_TIMEOUT_TICKS          = (5000/BSP_MS_PER_TICKS)
};

#define MAX_DS485_BACKOFF       4

/* Active object class -----------------------------------------------------*/

typedef struct DS485Tag {
    QActive super;
    QTimeEvt initialTimeEvt;
    QTimeEvt startupTimeEvt;
    QTimeEvt txMsgTimeEvt;
    QTimeEvt rxMsgTimeEvt;
    QTimeEvt joinResponseTimeEvt;
    QTimeEvt tokenMinTimeEvt;
    QTimeEvt tokenMaxTimeEvt;
    QTimeEvt findStationTimeEvt;
    QTimeEvt handOverTokenTimeEvt;
    QTimeEvt tokenMissingTimeEvt;
    QTimeEvt registerStationTimeEvt;
    QTimeEvt transmitDataTimeEvt;
    QEQueue deferredTxMessageQueue;
    dsid_t dsmid;
    uint8_t master;
    uint8_t deviceAddress;
    uint8_t successorDeviceAddress;
#ifdef DAEMON_IMPL
    dsuid_t busDsuids[DS485_MAX_SLAVE_ADDRESS+1];
    uint8_t dsidRequested[8];
#endif
    uint8_t usedAddresses[8];
    uint8_t pingDeviceAddress;
    uint8_t registerDeviceAddress;                   /**< cached device address for registration */
    uint16_t random;
    uint8_t try;
    uint8_t backoff;
    uint8_t findStation;
    uint8_t tokenMinHoldTimeElapsed;
    uint8_t tokenMaxHoldTimeElapsed;
    uint8_t tokenHandoverCount;
    uint8_t terminate;
    uint8_t messageSentAfterReceivingToken;
    uint8_t restartBus;
    QStateHandler superRegisterStationHistory;
    QStateHandler exitRegisterStationHistory;
    QStateHandler tryRegisterStationHistory;
    DS485Statistics statistics;
} DS485;

#include "unix/ds485-unix.h"
extern Ds485LibInst_t  devInst;

#ifdef DAEMON_IMPL
    static dsuid_t NO_DEVICE    = { { 0 } };
    dsuid_t ds485_dsid_to_dsuid(const dsid_t *dsid)
    {
        dsuid_t dsuid;

        dsuid.id[0] = dsid->id[0];
        dsuid.id[1] = dsid->id[1];
        dsuid.id[2] = dsid->id[2];
        dsuid.id[3] = dsid->id[3];
        dsuid.id[4] = dsid->id[4];
        dsuid.id[5] = dsid->id[5];
        dsuid.id[6] = 0;
        dsuid.id[7] = 0;
        dsuid.id[8] = 0;
        dsuid.id[9] = 0;
        dsuid.id[10] = dsid->id[6];
        dsuid.id[11] = dsid->id[7];
        dsuid.id[12] = dsid->id[8];
        dsuid.id[13] = dsid->id[9];
        dsuid.id[14] = dsid->id[10];
        dsuid.id[15] = dsid->id[11];
        dsuid.id[16] = 0;

        return dsuid;
    }
    static void dsuidToString(dsuid_t *dsuid, char *dsuid_str)
    {
        char *p;
        uint8_t i;
        p = dsuid_str;
        for (i = 0; i < sizeof(dsuid_t); ++i) {
            p += sprintf(p, "%02x", dsuid->id[i]);
        }
        dsuid_str[DSUID_STR_LEN - 1] = '\0';
    }
    static int cmpAddress( DS485 *me, uint8_t addr, dsuid_t* dsuid )
    {
        return  memcmp(&me->busDsuids[addr], dsuid, sizeof(dsuid_t));
    }
    static bool isAddressValid( DS485 *me, uint8_t addr )
    {
        return  cmpAddress(me,addr,&NO_DEVICE) != 0;
    }
    static void unregisterAddress( DS485 *me, uint8_t addr )
    {
        if (isAddressValid(me, addr))
        {
            char dsuid_str[DSUID_STR_LEN];
            dsuidToString(&me->busDsuids[addr], dsuid_str);
            ds485_system_msg(&devInst, LOG_WARNING, "Unregister dSUID %s as Busaddress %d\n", dsuid_str, addr);
            MMI_DS485DsuidRegisteredEvt *evt = Q_NEW( MMI_DS485DsuidRegisteredEvt, MMI_DS485_DSUID_DEREGISTRED_SIG );
            evt->dsuid = me->busDsuids[addr];
            QActive_postFIFO(AO_MMI, (QEvent*)evt);
            me->busDsuids[addr] = NO_DEVICE;
        }
    }
    static void registerAddress( DS485 *me, uint8_t addr, dsuid_t* dsuid )
    {
        if (!IsNullDsuid(*dsuid))
        {
            if (!IsEqualDsuid(*dsuid, me->busDsuids[addr])) {
                int i;
                bool reregister = FALSE;
                char dsuid_str[DSUID_STR_LEN];
                for (i = 0; i < sizeof(me->busDsuids)/sizeof(me->busDsuids[0]); i++) {
                    if (IsEqualDsuid(*dsuid, me->busDsuids[i])) {
                        dsuidToString(&me->busDsuids[i], dsuid_str);
                        ds485_system_msg(&devInst, LOG_WARNING, "Unregister dSUID %s as Busaddress %d\n", dsuid_str, i);
                        me->busDsuids[i] = NO_DEVICE;
                        reregister = TRUE;
                    } else if (addr == i) {
                        unregisterAddress(me, i);
                    }
                }

                me->busDsuids[addr] = *dsuid;
                dsuidToString(&me->busDsuids[addr], dsuid_str);
                ds485_system_msg(&devInst, LOG_WARNING, "Register dSUID %s as Busaddress %d\n", dsuid_str, addr);

                if (!reregister) {
                    MMI_DS485DsuidRegisteredEvt *evt = Q_NEW( MMI_DS485DsuidRegisteredEvt, MMI_DS485_DSUID_REGISTRED_SIG );
                    evt->dsuid = *dsuid;
                    QActive_postFIFO(AO_MMI, (QEvent*)evt);
                }
            }
        }
    }
    static void checkForMissingBusMembers( DS485 *me )
    {
        int i;
        for( i = 0; i < sizeof(me->busDsuids)/sizeof(me->busDsuids[0]); i++ ){
            if( IsBitSet8Cleared(me->usedAddresses,i) && isAddressValid( me, i ) ){
                unregisterAddress( me, i );
            }
        }
    }
#endif

static QState DS485_initial (DS485 *me, QEvent const *e);
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e);
static QState DS485_top (DS485 *me, QEvent const *e);
static QState DS485_startup(DS485 *me, QEvent const *e);
static QState DS485_send_dsid(DS485 *me, QEvent const *e);

static QState DS485_startup_slave(DS485 *me, QEvent const *e);
static QState DS485_wait_for_startup_request(DS485 *me, QEvent const *e);
static QState DS485_wait_for_join_request(DS485 *me, QEvent const *e);
static QState DS485_send_startup_response(DS485 *me, QEvent const *e);
static QState DS485_send_join_response(DS485 *me, QEvent const *e);
static QState DS485_wait_for_set_address_request(DS485 *me, QEvent const *e);
static QState DS485_wait_for_set_successor_address_request(DS485 *me, QEvent const *e);

static QState DS485_startup_master(DS485 *me, QEvent const *e);
static QState DS485_send_startup_request(DS485 *me, QEvent const *e);

static QState DS485_register_station(DS485 *me, QEvent const *e);
static QState DS485_ping_device_address(DS485 *me, QEvent const *e);
static QState DS485_send_set_address_request(DS485 *me, QEvent const *e);
static QState DS485_send_set_successor_address_request(DS485 *me, QEvent const *e);
static QState DS485_send_set_successor_address_request_to_predecessor(DS485 *me, QEvent const *e);

static QState DS485_operational(DS485 *me, QEvent const *e);

static QState DS485_holding_token(DS485 *me, QEvent const *e);
static QState DS485_find_station(DS485 *me, QEvent const *e);
static QState DS485_wait_for_token_min_hold_time(DS485 *me, QEvent const *e);
static QState DS485_transmit_data(DS485* me, QEvent const *e);
static QState DS485_wait_for_acknowledge(DS485* me, QEvent const *e);

static QState DS485_no_token(DS485 *me, QEvent const *e);
static QState DS485_hand_over_token(DS485 *me, QEvent const *e);
static QState DS485_send_token(DS485 *me, QEvent const *e);
static QState DS485_find_successor_device(DS485 *me, QEvent const *e);

static QState DS485_stopped(DS485 *me, QEvent const *e);

static void DS485_SendToken(uint8_t dest, uint8_t src);
static void DS485_SendMessage(DS485MessageEvt *msgEvt);
static void DS485_SendSetAddressResponse(DS485 *me, uint8_t dest);
static void DS485_SendSetSuccessorAddressResponse(DS485 *me, uint8_t dest);
static void DS485_PublishDeviceAddress(uint8_t deviceAddress);
static void DS485_ChangeToMaster(DS485 *me);
static uint8_t DS485_GetNextSuccessorAddress(DS485 *me, uint8_t startDeviceAddress);
static void DS485_ResetStatistics(DS485 *me);
static void DS485_RequestStatistics(DS485 *me, DS485ReqStatisticsEvt const * reqEvt);
#ifdef DAEMON_IMPL
static uint8_t DS485_getBusIdByDsuid( DS485 *me, dsuid_t *dsuid );
static dsuid_t DS485_getDsuidByBusId( DS485 *me, uint8_t busId );
#endif

/* Local objects -----------------------------------------------------------*/
static DS485 l_ds485;   /* the single instance of the dS485 active object */
static QEvent const * l_deferredTxMessageQueueSto[64];   /* event queue storage */
static QEvent const * l_ds485TxMessageQueueSto[8];    /* event queue storage */
static const QEvent l_ds485DeviceLostEvt = { MMI_DS485_DEVICE_LOST_SIG, 0 };
static QEQueue l_ds485TxMessageQueue;   /* transmit message queue */

/* Global-scope objects ----------------------------------------------------*/
QActive * const AO_DS485 = (QActive *)&l_ds485;      /* "opaque" AO pointer */

/*..........................................................................*/
void DS485_ctor(void)
{
    DS485 *me = &l_ds485;
    QActive_ctor(&me->super, (QStateHandler)&DS485_initial);
    /* initialize message event transmit queue */
    QEQueue_init(&l_ds485TxMessageQueue, l_ds485TxMessageQueueSto, Q_DIM(l_ds485TxMessageQueueSto));
    QEQueue_init(&me->deferredTxMessageQueue, l_deferredTxMessageQueueSto, Q_DIM(l_deferredTxMessageQueueSto));
    /* initialize time events */
    QTimeEvt_ctor(&me->initialTimeEvt, DS485_INITIAL_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->startupTimeEvt, DS485_STARTUP_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->txMsgTimeEvt, DS485_TX_MSG_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->rxMsgTimeEvt, DS485_RX_MSG_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->joinResponseTimeEvt, DS485_JOIN_RESPONSE_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->tokenMinTimeEvt, DS485_TOKEN_MIN_HOLD_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->tokenMaxTimeEvt, DS485_TOKEN_MAX_HOLD_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->findStationTimeEvt, DS485_FIND_STATION_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->handOverTokenTimeEvt, DS485_HAND_OVER_TOKEN_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->tokenMissingTimeEvt, DS485_TOKEN_MISSING_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->registerStationTimeEvt, DS485_REGISTER_STATION_TIMEOUT_SIG);
    QTimeEvt_ctor(&me->transmitDataTimeEvt, DS485_TRANSMIT_DATA_TIMEOUT_SIG);
    me->terminate = FALSE;
    me->backoff = 0;
    me->restartBus = FALSE;
}

void DS485_printQueueMargins(void)
{
    DS485 *me = &l_ds485;
    fprintf(stderr, "\tds485TxMessageQueue: %u/%lu\n", QEQueue_getNFree(&l_ds485TxMessageQueue), Q_DIM(l_ds485TxMessageQueueSto));
    fprintf(stderr, "\tdeferredTxMessageQueue: %u/%lu\n", QEQueue_getNFree(&me->deferredTxMessageQueue), Q_DIM(l_deferredTxMessageQueueSto));
}

/**
 * Create DS485 control message
 */
DS485MessageEvt * DS485_CreateControlMessage(uint8_t dest, uint8_t src, DS485CommandType type, DS485CommandId cmd, uint8_t payloadLength)
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_SHORT_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
    message[DS485_DEST_OFFSET] = dest | DS485_TYPE_CONTROL;
    message[DS485_SRC_OFFSET] = src | type;
    message[DS485_CL_OFFSET] = (cmd << 4) | (payloadLength & 0xF);
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
    return msgEvt;
}

/**
 * Create DS485 data message
 */
#ifdef DAEMON_IMPL
DS485MessageEvt * DS485_CreateDataMessage(const dsuid_t *dest, DS485CommandType type, uint8_t payloadLength)
#else
DS485MessageEvt * DS485_CreateDataMessage(uint8_t dest, uint8_t src, DS485CommandType type, uint8_t payloadLength)
#endif
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_SHORT_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485MessageEvt * msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
#ifdef DAEMON_IMPL
    message[DS485_DEST_OFFSET]  = DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET]   = 0;
#else
    message[DS485_DEST_OFFSET] = dest | DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET] = src | type;
#endif
    message[DS485_CL_OFFSET] = payloadLength;
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
#ifdef DAEMON_IMPL
    msgEvt->destinationAddr  = *dest;
#endif
    return msgEvt;
}

/**
 * Create long DS485 data message
 */
#ifdef DAEMON_IMPL
DS485LongMessageEvt * DS485_CreateLongDataMessage(const dsuid_t *dest, DS485CommandType type, uint8_t payloadLength)
#else
DS485LongMessageEvt * DS485_CreateLongDataMessage(uint8_t dest, uint8_t src, DS485CommandType type, uint8_t payloadLength)
#endif
{
    uint8_t* message;
    Q_ASSERT(payloadLength <= DS485_LONG_PAYLOAD_SIZE);
    /* allocate new message event */
    DS485LongMessageEvt * msgEvt = Q_NEW(DS485LongMessageEvt, DS485_TX_MSG_SIG);
    /* build header */
    message = msgEvt->message;
    message[DS485_START_OFFSET] = DS485_START_BYTE;
#ifdef DAEMON_IMPL
    message[DS485_DEST_OFFSET]  = DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET]   = 0;
#else
    message[DS485_DEST_OFFSET] = dest | DS485_TYPE_DATA;
    message[DS485_SRC_OFFSET] = src | type;
#endif
    message[DS485_CL_OFFSET] = payloadLength;
    /* set message size = message length + 4 header + 2 CRC-16 */
    msgEvt->size = payloadLength + DS485_HEADER_CRC_SIZE;
#ifdef DAEMON_IMPL
    msgEvt->destinationAddr  = *dest;
#endif
    return msgEvt;
}

/**
 * Return next DS485 message to transmit or null
 */
DS485MessageEvt const * DS485_GetNextTxMessage(void)
{
    return (DS485MessageEvt const *)QEQueue_get(&l_ds485TxMessageQueue);
}

/*..........................................................................*/
static QState DS485_initial(DS485 *me, QEvent const *e)
{
    (void)e;        /* suppress the compiler warning about unused parameter */
    /* QS dictionary entries */
    QS_OBJ_DICTIONARY(&l_ds485);

    QS_FUN_DICTIONARY(&DS485_initial);
    QS_FUN_DICTIONARY(&DS485_wait_for_configuration);
    QS_FUN_DICTIONARY(&DS485_top);
    QS_FUN_DICTIONARY(&DS485_startup);
    QS_FUN_DICTIONARY(&DS485_send_dsid);

    QS_FUN_DICTIONARY(&DS485_startup_slave);
    QS_FUN_DICTIONARY(&DS485_wait_for_join_request);
    QS_FUN_DICTIONARY(&DS485_wait_for_startup_request);
    QS_FUN_DICTIONARY(&DS485_send_join_response);
    QS_FUN_DICTIONARY(&DS485_wait_for_set_address_request);
    QS_FUN_DICTIONARY(&DS485_wait_for_set_successor_address_request);

    QS_FUN_DICTIONARY(&DS485_startup_master);
    QS_FUN_DICTIONARY(&DS485_send_startup_request);

    QS_FUN_DICTIONARY(&DS485_register_station);
    QS_FUN_DICTIONARY(&DS485_ping_device_address);
    QS_FUN_DICTIONARY(&DS485_send_set_address_request);
    QS_FUN_DICTIONARY(&DS485_send_set_successor_address_request);
    QS_FUN_DICTIONARY(&DS485_send_set_successor_address_request_to_predecessor);

    QS_FUN_DICTIONARY(&DS485_operational);

    QS_FUN_DICTIONARY(&DS485_holding_token);
    QS_FUN_DICTIONARY(&DS485_find_station);
    QS_FUN_DICTIONARY(&DS485_wait_for_token_min_hold_time);
    QS_FUN_DICTIONARY(&DS485_transmit_data);
    QS_FUN_DICTIONARY(&DS485_wait_for_acknowledge);

    QS_FUN_DICTIONARY(&DS485_no_token);
    QS_FUN_DICTIONARY(&DS485_hand_over_token);
    QS_FUN_DICTIONARY(&DS485_send_token);

    QS_FUN_DICTIONARY(&DS485_stopped);

    QS_SIG_DICTIONARY(DS485_STARTUP_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_TX_MSG_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_JOIN_RESPONSE_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_TOKEN_MIN_HOLD_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_TOKEN_MAX_HOLD_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_FIND_STATION_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_HAND_OVER_TOKEN_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_TOKEN_MISSING_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_REGISTER_STATION_TIMEOUT_SIG, 0);
    QS_SIG_DICTIONARY(DS485_TRANSMIT_DATA_TIMEOUT_SIG, 0);

    /* subscribe events */
    QActive_subscribe((QActive *)me, CONFIG_INIT_SIG);
    /* initialize the USART3 */
#ifndef DAEMON_IMPL
    USART3_Init();
    /*---> system_uart_init(); */
#endif
    return Q_TRAN(&DS485_wait_for_configuration);
}

/*..........................................................................*/
static QState DS485_wait_for_configuration(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case CONFIG_INIT_SIG: {
        DatamodelConfigEvt const * config = (DatamodelConfigEvt const *)e;
        /* save dSM ID */
        me->dsmid = config->dsmid;
        /* subscribe events */
        QActive_subscribe((QActive *)me, SYSTEM_TERM_SIG);
        return Q_TRAN(&DS485_startup);
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static QState DS485_top(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TX_MSG_SIG:
        /* defer transmit message event */
        if (QEQueue_getNFree(&me->deferredTxMessageQueue) != 0) {
            QActive_defer((QActive *)me, &me->deferredTxMessageQueue, e);
        }
        return Q_HANDLED();
    case DS485_REQ_STATISTICS_SIG:
        DS485_RequestStatistics(me, (DS485ReqStatisticsEvt const *)e);
        return Q_HANDLED();
    case DS485_RESET_STATISTICS_SIG:
        DS485_ResetStatistics(me);
        return Q_HANDLED();
    case SYSTEM_TERM_SIG:
        return Q_TRAN(&DS485_stopped);
#ifdef DAEMON_IMPL
    case DS485_REQ_BUS_MEMBERS_SIG:
    {
        DS485_DsuidsResponseEvt *evt = Q_NEW(DS485_DsuidsResponseEvt, DS485_RES_BUS_MEMBERS_SIG);
        memset( evt->dsuids, 0, sizeof(evt->dsuids) );
        ccSendResponse( e, (QEvent*)evt );
        return Q_HANDLED();
    }
    case DS485_TX_RESTART_BUS_SIG:
    {
        postTxResponse( 0 );

        if (me->master) {
            me->restartBus = TRUE;
        }
        DS485MessageEvt * evt = DS485_CreateControlMessage(0, me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_RESTART_BUS, 0);
        QActive_postFIFO(AO_DS485, (QEvent *) evt);

        return Q_HANDLED();
    }
#endif
    }
    return Q_SUPER(&QHsm_top);
}
/*..........................................................................*/
static QState DS485_startup(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* generate a random time between 2000 and 4000 ms in steps of 4 ms */
        uint16_t time = (2000 + ((rand() % 500 + 1) * 4)) * (1 << me->backoff);
        QTimeEvt_postIn(&me->initialTimeEvt, (QActive *)me, (time / BSP_MS_PER_TICKS));
        /* increment exponential back-off for startup timeout */
        if (me->backoff < MAX_DS485_BACKOFF) {
            ++me->backoff;
        }
        /* reset used addresses */
        memset(me->usedAddresses, 0, sizeof(me->usedAddresses));
#ifdef DAEMON_IMPL
        checkForMissingBusMembers(me);
        memset(me->dsidRequested, 0, sizeof(me->dsidRequested));
#endif
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->initialTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_TOKEN_SIG:
        /* fall through */
    case DS485_RX_STARTUP_SIG:
        /* fall through */
    case DS485_RX_JOIN_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_DATA_SIG: {
        /* any message received -> startup as slave */
        return Q_TRAN(&DS485_startup_slave);
    }
    case DS485_INITIAL_TIMEOUT_SIG: {
        return Q_TRAN(&DS485_send_dsid);
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_send_dsid(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send startup request with own dSID */
        DS485MessageEvt* txMsgEvt = DS485_CreateControlMessage(0, 0, DS485_TYPE_REQ, DS485_CMD_STARTUP, sizeof(me->dsmid));
        memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txStartupRequest;
        QTimeEvt_postIn(&me->txMsgTimeEvt, (QActive *)me, TX_MSG_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->txMsgTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        QTimeEvt_disarm(&me->txMsgTimeEvt);
        if (DS485_IsRequest(rxMsgEvt)) {
            /* compare dSID */
            if (memcmp(&rxMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid)) == 0) {
                /* startup request with own dSID received -> start up as master */
                return Q_TRAN(&DS485_startup_master);
            }
        }
        /* startup request with another dSID or startup response received -> start up as slave */
        return Q_TRAN(&DS485_startup_slave);
    }
    case DS485_RX_TOKEN_SIG:
        /* fall through */
    case DS485_RX_JOIN_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_DATA_SIG: {
        /* any message received -> startup as slave */
        return Q_TRAN(&DS485_startup_slave);
    }
    case DS485_TX_MSG_TIMEOUT_SIG: {
        /* timeout occurred -> restart from scratch */
        return Q_TRAN(&DS485_startup);
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_startup_slave(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->master = FALSE;
        QTimeEvt_postIn(&me->startupTimeEvt, (QActive *)me, SLAVE_STARTUP_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->startupTimeEvt);
        return Q_HANDLED();
    }
    case DS485_STARTUP_TIMEOUT_SIG: {
        /* timeout occurred -> restart from scratch */
        return Q_TRAN(&DS485_startup);
    }
    case DS485_RX_STARTUP_SIG: {
        return Q_TRAN(&DS485_wait_for_startup_request);
    }
    case DS485_RX_JOIN_SIG: {
        return Q_TRAN(&DS485_wait_for_join_request);
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_wait_for_startup_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* generate a random number between 1 and 64 */
        me->random = (rand() % 64) + 1;
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        if (DS485_IsRequest((DS485MessageEvt * const)e)) {
            if (--me->random == 0) {
                return Q_TRAN(&DS485_send_startup_response);
            }
        }
        return Q_HANDLED();
    }
    case DS485_RX_JOIN_SIG: {
        if (DS485_IsRequest((DS485MessageEvt * const)e)) {
            return Q_TRAN(&DS485_wait_for_join_request);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_join_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* generate a random number between 1 and 10 */
        me->random = (rand() % 10) + 1;
        return Q_HANDLED();
    }
    case DS485_RX_JOIN_SIG: {
        if (DS485_IsRequest((DS485MessageEvt * const)e)) {
            if (--me->random == 0) {
                return Q_TRAN(&DS485_send_join_response);
            }
        }
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        if (DS485_IsRequest((DS485MessageEvt * const)e)) {
            return Q_TRAN(&DS485_wait_for_startup_request);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_send_startup_response(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send startup response */
        DS485MessageEvt* txMsgEvt = DS485_CreateControlMessage(0, 0, DS485_TYPE_ACK_RES, DS485_CMD_STARTUP, sizeof(me->dsmid));
        memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txStartupResponse;
        QTimeEvt_postIn(&me->txMsgTimeEvt, (QActive *)me, WAIT_FOR_STARTUP_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->txMsgTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            if (memcmp(&rxMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid)) == 0) {
                /* set intermediate address */
                me->deviceAddress = DS485_INTERMEDIATE_ADDRESS;
                return Q_TRAN(&DS485_wait_for_set_address_request);
            }
        }
        return Q_TRAN(&DS485_startup_slave);
    }
    case DS485_TX_MSG_TIMEOUT_SIG: {
        /* timeout occurred */
        return Q_TRAN(&DS485_startup_slave);
    }
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_send_join_response(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send join response */
        DS485MessageEvt* txMsgEvt = DS485_CreateControlMessage(0, 0, DS485_TYPE_ACK_RES, DS485_CMD_JOIN, sizeof(me->dsmid));
        memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txJoinResponse;
        QTimeEvt_postIn(&me->txMsgTimeEvt, (QActive *)me, WAIT_FOR_JOIN_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->txMsgTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_JOIN_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            if (memcmp(&rxMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid)) == 0) {
                /* set intermediate address */
                me->deviceAddress = DS485_INTERMEDIATE_ADDRESS;
                return Q_TRAN(&DS485_wait_for_set_address_request);
            }
        }
        return Q_TRAN(&DS485_startup_slave);
    }
    case DS485_TX_MSG_TIMEOUT_SIG: {
        /* timeout occurred */
        return Q_TRAN(&DS485_startup_slave);
    }
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_set_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvt_postIn(&me->rxMsgTimeEvt, (QActive *)me, WAIT_FOR_SET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->rxMsgTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_SET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            dsuid_t dsuid;
            /* save my device address */
            me->deviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set device address response */
            DS485_SendSetAddressResponse(me, DS485_GetSource(rxMsgEvt));
            dsuid = ds485_dsid_to_dsuid(&me->dsmid);
            registerAddress(me, me->deviceAddress, &dsuid);
            /* wait for set successor address request */
            return Q_TRAN(&DS485_wait_for_set_successor_address_request);
        }
        return Q_HANDLED();
    }
    case DS485_RX_JOIN_SIG: {
        return Q_TRAN(&DS485_wait_for_join_request);
    }
    case DS485_RX_STARTUP_SIG: {
        return Q_TRAN(&DS485_wait_for_startup_request);
    }
    case DS485_RX_MSG_TIMEOUT_SIG: {
        return Q_TRAN(&DS485_startup_slave);
    }
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_wait_for_set_successor_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        QTimeEvt_postIn(&me->rxMsgTimeEvt, (QActive *)me, WAIT_FOR_SET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->rxMsgTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt) && (DS485_GetDestination(rxMsgEvt) == me->deviceAddress)) {
            /* save successor device address */
            me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set successor address response */
            DS485_SendSetSuccessorAddressResponse(me, DS485_GetSource(rxMsgEvt));

            SetBitSet8(me->usedAddresses, me->deviceAddress);
            /* go to operational state */
            return Q_TRAN(&DS485_operational);
        }
        return Q_HANDLED();
    }
    case DS485_RX_SET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            dsuid_t dsuid;
            /* save my device address */
            me->deviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set address response */
            DS485_SendSetAddressResponse(me, DS485_GetSource(rxMsgEvt));
            dsuid = ds485_dsid_to_dsuid(&me->dsmid);
            registerAddress(me, me->deviceAddress, &dsuid);
            /* wait for set successor address request */
            return Q_TRAN(&DS485_wait_for_set_successor_address_request);
        }
        return Q_HANDLED();
    }
    case DS485_RX_MSG_TIMEOUT_SIG: {
        return Q_TRAN(&DS485_startup_slave);
    }
    }
    return Q_SUPER(&DS485_startup_slave);
}

/*..........................................................................*/
static QState DS485_startup_master(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->master = TRUE;
        SetBitSet8(me->usedAddresses, DS485_MASTER_ADDRESS);
        me->deviceAddress = DS485_MASTER_ADDRESS;
#ifdef DAEMON_IMPL
        dsuid_t dsuid = ds485_dsid_to_dsuid(&me->dsmid);
        registerAddress(me, me->deviceAddress, &dsuid);
#endif
        me->random = 0;
        me->restartBus = FALSE;
        return Q_HANDLED();
    }
    case Q_INIT_SIG: {
        return Q_TRAN(&DS485_send_startup_request);
    }
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt) && (DS485_GetDestination(rxMsgEvt) == me->deviceAddress)) {
            /* save successor device address */
            me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
            /* send set successor address response */
            DS485_SendSetSuccessorAddressResponse(me, DS485_GetSource(rxMsgEvt));
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_send_startup_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* increment counter */
        ++me->random;
        /* send startup request */
        txMsgEvt = DS485_CreateControlMessage(0, 0, DS485_TYPE_REQ, DS485_CMD_STARTUP, sizeof(me->dsmid));
        memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txStartupRequest;
        QTimeEvt_postIn(&me->startupTimeEvt, (QActive *)me, STARTUP_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->startupTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_STARTUP_SIG: {
        DS485MessageEvt * const rxMsgEvt = (DS485MessageEvt * const)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            if (memcmp(&rxMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, 12) == 0) {
                /* skip own startup request */
                return Q_HANDLED();
            }
            /* there is another master on the bus */
            return Q_TRAN(&DS485_startup);
        }
        /* store history */
        me->superRegisterStationHistory = (QStateHandler)&DS485_startup_master;
        me->exitRegisterStationHistory = (QStateHandler)&DS485_send_startup_request;
        return Q_TRAN(&DS485_register_station);
    }
    case DS485_STARTUP_TIMEOUT_SIG: {
        if (me->random < 256) {
            return Q_TRAN(&DS485_send_startup_request);
        }
        return Q_TRAN(&DS485_operational);
    }
    }
    return Q_SUPER(&DS485_startup_master);
}

/*..........................................................................*/
static QState DS485_register_station(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        me->try = 0;
        ++me->statistics.startRegisterStation;
        return Q_HANDLED();
    }
    case Q_INIT_SIG: {
        /* get first slave device address to ping */
        me->pingDeviceAddress = DS485_MIN_SLAVE_ADDRESS;
        return Q_TRAN(&DS485_ping_device_address);
    }
    case DS485_REGISTER_STATION_TIMEOUT_SIG: {
        if (me->try < 2) {
            /* retry again -> go to try history */
            return Q_TRAN(me->tryRegisterStationHistory);
        }
        /* exit to history */
        if (me->restartBus) {
            return Q_TRAN(&DS485_startup_master);
        }
        return Q_TRAN(me->exitRegisterStationHistory);
    }
    }
    return Q_SUPER(me->superRegisterStationHistory);
}

/*..........................................................................*/
static QState DS485_ping_device_address(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send get address request (ping) */
        DS485_SendMessage(DS485_CreateControlMessage(me->pingDeviceAddress, me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_GET_ADDRESS, 0));
        ++me->statistics.txGetAddressRequest;
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, GET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* stop timer */
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    }
    case DS485_REGISTER_STATION_TIMEOUT_SIG: {
        /* get address request (ping) timed out */
        /* reset try counter */
        me->try = 0;
        me->registerDeviceAddress = me->pingDeviceAddress;
        return Q_TRAN(&DS485_send_set_address_request);
    }
    case DS485_RX_GET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            /* get address response received */
            if (DS485_GetSource(rxMsgEvt) == me->pingDeviceAddress) {
                /* we got a ping response -> set address to used */
                SetBitSet8(me->usedAddresses, me->pingDeviceAddress);
                /* get next slave device address to ping */
                ++me->pingDeviceAddress;
                if (me->pingDeviceAddress <= DS485_MAX_SLAVE_ADDRESS) {
                    return Q_TRAN(&DS485_ping_device_address);
                }
                /* all addresses used -> exit to history */
                if (me->restartBus) {
                    return Q_TRAN(&DS485_startup_master);
                }
                return Q_TRAN(me->exitRegisterStationHistory);
            }
        }
        break;
    }
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_send_set_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* store history */
        me->tryRegisterStationHistory = (QStateHandler)&DS485_send_set_address_request;
        /* increment try counter */
        ++me->try;
        /* send set address request */
        txMsgEvt = DS485_CreateControlMessage(DS485_INTERMEDIATE_ADDRESS, me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_SET_ADDRESS, 1);
        txMsgEvt->message[DS485_DATA_OFFSET] = me->registerDeviceAddress;
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txSetAddressRequest;
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, SET_ADDRESS_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_SET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            if (DS485_GetSource(rxMsgEvt) == me->registerDeviceAddress) {
                /* reset try counter */
                me->try = 0;
                /* set successor address request */
                return Q_TRAN(&DS485_send_set_successor_address_request);
            }
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_send_set_successor_address_request(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* store history */
        me->tryRegisterStationHistory = (QStateHandler)&DS485_send_set_successor_address_request;
        /* increment try counter */
        ++me->try;
        /* send set successor address request */
        txMsgEvt = DS485_CreateControlMessage(me->registerDeviceAddress, me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_SET_SUCCESSOR_ADDRESS, 1);
        /* get successor address */
        txMsgEvt->message[DS485_DATA_OFFSET] = DS485_GetNextSuccessorAddress(me, me->registerDeviceAddress);
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txSetSuccessorAddressRequest;
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, SET_ADDRESS_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            /* check successor address response */
            if ((DS485_GetDestination(rxMsgEvt) == me->deviceAddress) && (DS485_GetSource(rxMsgEvt) == me->registerDeviceAddress)) {
                /* disarm register station timer */
                QTimeEvt_disarm(&me->registerStationTimeEvt);
                /* reset try counter */
                me->try = 0;
                return Q_TRAN(&DS485_send_set_successor_address_request_to_predecessor);
            }
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_send_set_successor_address_request_to_predecessor(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        /* set history to exit */
        me->tryRegisterStationHistory = me->exitRegisterStationHistory;
        /* increment try counter */
        ++me->try;
        /* send set successor address request */
        txMsgEvt = DS485_CreateControlMessage((me->registerDeviceAddress - 1), me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_SET_SUCCESSOR_ADDRESS, 1);
        txMsgEvt->message[DS485_DATA_OFFSET] = me->registerDeviceAddress;
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txSetSuccessorAddressRequest;
        /* start timer */
        QTimeEvt_postIn(&me->registerStationTimeEvt, (QActive *)me, SET_ADDRESS_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->registerStationTimeEvt);
        return Q_HANDLED();
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
                /* this device is the successor */
                me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
                /* send set successor address response */
                DS485_SendSetSuccessorAddressResponse(me, DS485_GetSource(rxMsgEvt));
            }
            return Q_HANDLED();
        } else if (DS485_IsResponse(rxMsgEvt)) {
            if ((DS485_GetDestination(rxMsgEvt) == me->deviceAddress) && (DS485_GetSource(rxMsgEvt) == (me->registerDeviceAddress - 1))) {
                static const QEvent ds485DeviceRegistered = { MMI_DS485_DEVICE_REGISTERED_SIG, 0 };
                /* disarm register station timer */
                QTimeEvt_disarm(&me->registerStationTimeEvt);
                /* everything done -> set registered device address used */
                SetBitSet8(me->usedAddresses, me->registerDeviceAddress);
                /* send device registered to MMI */
                QActive_postFIFO(AO_MMI, &ds485DeviceRegistered);
                ++me->statistics.doneRegisterStation;
                /* exit to history */
                if (me->restartBus) {
                    return Q_TRAN(&DS485_startup_master);
                }
                return Q_TRAN(me->exitRegisterStationHistory);
            }
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&DS485_register_station);
}

/*..........................................................................*/
static QState DS485_operational(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
#ifdef DAEMON_IMPL
    case DS485_RX_GET_ADDRESS_SIG:
    {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse((DS485MessageEvt const *)e)) {
            if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
                dsuid_t dsuid = ds485_dsid_to_dsuid((dsid_t*)&rxMsgEvt->message[DS485_DATA_OFFSET]);
                registerAddress(me, DS485_GetSource(rxMsgEvt), &dsuid);
                ClearBitSet8(me->dsidRequested, DS485_GetSource(rxMsgEvt));
            }
        }
        return Q_HANDLED();
    }
#endif
    case Q_ENTRY_SIG: {
        /* reset exponential back-off for startup timeout */
        me->backoff = 0;
        /* publish device address */
        DS485_PublishDeviceAddress(me->deviceAddress);
        /* notify ready to MMI */
        static const MMIDS485StateEvt readyEvt = { { MMI_DS485_STATE_SIG, 0 }, MMI_DS485_READY };
        QActive_postFIFO(AO_MMI, (QEvent *)&readyEvt);
        ++me->statistics.enterOperational;
        /* initialize find station flag */
        QTimeEvt_disarm(&me->findStationTimeEvt);
        me->findStation = me->master;
#ifdef DAEMON_IMPL
        memset(me->dsidRequested, 0, sizeof(me->dsidRequested));
#endif
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        static const MMIDS485StateEvt initializingEvt = { { MMI_DS485_STATE_SIG, 0 }, MMI_DS485_INITIALIZING };
        QActive_postFIFO(AO_MMI, (QEvent *)&initializingEvt);
        return Q_HANDLED();
    }
    case Q_INIT_SIG:
        if (me->master) {
            return Q_TRAN(&DS485_holding_token);
        }
        return Q_TRAN(&DS485_no_token);
    case DS485_FIND_STATION_TIMEOUT_SIG:
        me->findStation = me->master;
        return Q_HANDLED();
    case SYSTEM_TERM_SIG:
        me->terminate = TRUE;
        return Q_HANDLED();
    case DS485_RX_JOIN_SIG:
        /* fall through */
    case DS485_RX_STARTUP_SIG: {
        /* only master is allowed to send join and startup requests */
        if (me->master) {
            DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
            if (DS485_IsRequest(rxMsgEvt)) {
                if (memcmp(&rxMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, 12) != 0) {
                    return Q_TRAN(&DS485_startup);
                }
            }
        }
        return Q_HANDLED();
    }
#ifdef DAEMON_IMPL
    case DS485_REQ_BUS_MEMBERS_SIG:
    {
        DS485_DsuidsResponseEvt *evt = Q_NEW(DS485_DsuidsResponseEvt, DS485_RES_BUS_MEMBERS_SIG);
        memcpy( evt->dsuids, me->busDsuids, sizeof(evt->dsuids) );
        ccSendResponse( e, (QEvent*)evt );
        return Q_HANDLED();
    }
#endif
    }
    return Q_SUPER(&DS485_top);
}

/*..........................................................................*/
static QState DS485_holding_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        me->tokenMinHoldTimeElapsed = FALSE;
        me->tokenMaxHoldTimeElapsed = FALSE;
        /* start minimum hold timeout only if there are other devices on the bus */
        if ((me->deviceAddress == 0) && (me->successorDeviceAddress == 0)) {
            QTimeEvt_postIn(&me->tokenMaxTimeEvt, (QActive *)me, TOKEN_LONG_HOLD_TIMEOUT_TICKS);
        } else {
            QTimeEvt_postIn(&me->tokenMinTimeEvt, (QActive *)me, TOKEN_MIN_HOLD_TIMEOUT_TICKS);
            QTimeEvt_postIn(&me->tokenMaxTimeEvt, (QActive *)me, TOKEN_MAX_HOLD_TIMEOUT_TICKS);
        }
        me->messageSentAfterReceivingToken = FALSE;
        return Q_HANDLED();
    case Q_EXIT_SIG:
        /* disarm timers */
        QTimeEvt_disarm(&me->tokenMinTimeEvt);
        QTimeEvt_disarm(&me->tokenMaxTimeEvt);
        return Q_HANDLED();
    case Q_INIT_SIG: {
        /* check for transmit data ready (internal message queue) */
        if (QActive_recall((QActive *)me, &me->deferredTxMessageQueue)) {
            return Q_TRAN(&DS485_transmit_data);
        }
        if (me->findStation) {
            return Q_TRAN(&DS485_find_station);
        }
        return Q_TRAN(&DS485_wait_for_token_min_hold_time);
    }
    case DS485_TOKEN_MIN_HOLD_TIMEOUT_SIG:
        me->tokenMinHoldTimeElapsed = TRUE;
        return Q_HANDLED();
    case DS485_TOKEN_MAX_HOLD_TIMEOUT_SIG:
        me->tokenMaxHoldTimeElapsed = TRUE;
        return Q_HANDLED();
    }
    return Q_SUPER(&DS485_operational);
}

/*..........................................................................*/
static QState DS485_transmit_data(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TX_MSG_SIG: {
        bool isRequest = FALSE;
        /* NOTE: cast away const! */
        DS485LongMessageEvt * txMsgEvt = (DS485LongMessageEvt *)e;
        /* set message sent after receiving token flag */
        me->messageSentAfterReceivingToken = TRUE;
        /* update source address */
        DS485_SetSource(txMsgEvt, me->deviceAddress);
#ifdef DAEMON_IMPL
        if( DS485_IsData( txMsgEvt ) ){
            uint8_t dst;
            if( IsBroadcastDsuid(txMsgEvt->destinationAddr) ){
                dst   = DS485_BROADCAST_MASK;
            }
            else {
                dst = DS485_getBusIdByDsuid( me, &txMsgEvt->destinationAddr );
                if( dst == DS485_INVALID_ADDRESS ){
                    postTxResponse( DS485_INTERNAL_ERROR_INVALID_BUS_ADDR );
                    /* recall next event */
                    if (QActive_recall((QActive *)me, &me->deferredTxMessageQueue)) {
                        return Q_HANDLED();
                    }
                    /* no more messages */
                    return Q_TRAN(&DS485_wait_for_token_min_hold_time);
                }
            }
            DS485_SetDestination( txMsgEvt, dst );
        }
#endif
        /* send message */
        DS485_SendMessage((DS485MessageEvt*)txMsgEvt);
#ifdef DAEMON_IMPL
        if( DS485_IsData( txMsgEvt ) ){
            postTxResponse( 0 );
        }
#endif
        switch (DS485_GetCommandType(txMsgEvt)) {
        case DS485_TYPE_REQ:
            ++me->statistics.txDataRequest;
            isRequest = TRUE;
            break;
        case DS485_TYPE_ACK_RES:
            ++me->statistics.txDataResponse;
            break;
        case DS485_TYPE_EVENT:
            ++me->statistics.txDataEvent;
            break;
        default:
            break;
        }
        /* is message broadcast or not a request (i.e. a response)? */
        if (DS485_IsBroadcast(txMsgEvt) || !isRequest) {
            /* recall next event */
            if (QActive_recall((QActive *)me, &me->deferredTxMessageQueue)) {
                return Q_HANDLED();
            }
            /* no more messages to transmit */
            if (me->restartBus) {
                return Q_TRAN(&DS485_startup_master);
            }
            /* is minimum hold time elapsed? */
            if (me->tokenMinHoldTimeElapsed) {
                return Q_TRAN(&DS485_hand_over_token);
            }
            return Q_TRAN(&DS485_wait_for_token_min_hold_time);

        }
        /* message is a unicast request -> wait for acknowledge */
        return Q_TRAN(&DS485_wait_for_acknowledge);
    }
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_wait_for_acknowledge(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* start response timer */
        QTimeEvt_postIn(&me->transmitDataTimeEvt, (QActive *)me, DSMAPI_RESPONSE_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        QTimeEvt_disarm(&me->transmitDataTimeEvt);
        return Q_HANDLED();
    }
    case DS485_RX_DATA_SIG: {
        switch (DS485_GetCommandType((DS485MessageEvt const *)e)) {
        case DS485_TYPE_REQ:
            /* fall through */
        case DS485_TYPE_EVENT: {
            /* just ignore incoming requests and events */
            return Q_HANDLED();
        }
        default: {
            break;
        }
        }
        /* fall through */
    }
    case DS485_RX_RESPONSE_SIG: {
#ifdef DAEMON_IMPL
        DS485MessageEvt * rxMsgEvt = (DS485MessageEvt *)e;
        rxMsgEvt->sourceAddr = DS485_getDsuidByBusId(me, DS485_GetSource(rxMsgEvt));
        /* responses are not sent to broadcast */
        rxMsgEvt->destinationAddr = DS485_getDsuidByBusId(me, DS485_GetDestination(rxMsgEvt));
        /* post message to DSM API active object */
        QActive_postFIFO(AO_DsmApi, e);
#else
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
#endif
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* stop response timer */
            QTimeEvt_disarm(&me->transmitDataTimeEvt);
            /* update statistics */
            if (e->sig == DS485_RX_RESPONSE_SIG) {
                ++me->statistics.rxResponse;
            } else {
                ++me->statistics.rxDataResponse;
            }
#ifndef DAEMON_IMPL
            /* post message to DSM API active object */
            QActive_postFIFO(AO_DsmApi, e);
#endif
            if (me->restartBus) {
                return Q_TRAN(&DS485_startup_master);
            }
            /* is maximum hold time elapsed? */
            if (me->tokenMaxHoldTimeElapsed) {
                return Q_TRAN(&DS485_hand_over_token);
            }
            /* recall next event */
            if (QActive_recall((QActive *)me, &me->deferredTxMessageQueue)) {
                return Q_TRAN(&DS485_transmit_data);
            }
            /* no more message to transmit */
            /* is minimum hold time elapsed? */
            if (me->tokenMinHoldTimeElapsed) {
                return Q_TRAN(&DS485_hand_over_token);
            }
            return Q_TRAN(&DS485_wait_for_token_min_hold_time);
        }
        return Q_HANDLED();
    }
#ifdef DAEMON_IMPL
    case DS485_RX_GET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
                dsuid_t dsuid = ds485_dsid_to_dsuid((dsid_t*)&rxMsgEvt->message[DS485_DATA_OFFSET]);
                registerAddress(me, DS485_GetSource(rxMsgEvt), &dsuid);
                ClearBitSet8(me->dsidRequested, DS485_GetSource(rxMsgEvt));
                if (me->restartBus) {
                    return Q_TRAN(&DS485_startup_master);
                }
                /* max hold time elapsed? */
                if (me->tokenMaxHoldTimeElapsed) {
                    return Q_TRAN(&DS485_hand_over_token);
                }
                /* recall next event */
                if (QActive_recall((QActive *)me, &me->deferredTxMessageQueue)) {
                    return Q_TRAN(&DS485_transmit_data);
                }
                /* min hold time elapsed? */
                if (me->tokenMinHoldTimeElapsed) {
                    return Q_TRAN(&DS485_hand_over_token);
                }
                return Q_TRAN(&DS485_wait_for_token_min_hold_time);
            }
        }
        break;
    }
#endif
    case DS485_TRANSMIT_DATA_TIMEOUT_SIG: {
        if (me->restartBus) {
            return Q_TRAN(&DS485_startup_master);
        }
        /* timed out -> hand over token */
        return Q_TRAN(&DS485_hand_over_token);
    }
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_find_station(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        DS485MessageEvt* txMsgEvt;
        me->findStation = FALSE;
        /* send join request */
        txMsgEvt = DS485_CreateControlMessage(0, me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_JOIN, sizeof(me->dsmid));
        memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
        DS485_SendMessage(txMsgEvt);
        ++me->statistics.txJoinRequest;
        QTimeEvt_postIn(&me->joinResponseTimeEvt, (QActive *)me, JOIN_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    }
    case Q_EXIT_SIG: {
        /* stop response timer */
        QTimeEvt_disarm(&me->joinResponseTimeEvt);
        /* restart find station timeout */
#ifdef DAEMON_IMPL
        if (me->successorDeviceAddress == 0)
           QTimeEvt_postIn(&me->findStationTimeEvt, (QActive *)me, FIND_STATION_SECOND_DEVICE_TIMEOUT_TICKS);
        else
#endif
        QTimeEvt_postIn(&me->findStationTimeEvt, (QActive *)me, FIND_STATION_TIMEOUT_TICKS);
    }
    case DS485_RX_JOIN_SIG: {
        if (DS485_IsResponse((DS485MessageEvt * const)e)) {
            QTimeEvt_disarm(&me->joinResponseTimeEvt);
            me->superRegisterStationHistory = (QStateHandler)&DS485_holding_token;
            me->exitRegisterStationHistory = (QStateHandler)&DS485_hand_over_token;
            return Q_TRAN(&DS485_register_station);
        }
        return Q_HANDLED();
    }
    case DS485_JOIN_RESPONSE_TIMEOUT_SIG: {
        if (me->restartBus) {
            return Q_TRAN(&DS485_startup_master);
        }
        return Q_TRAN(&DS485_hand_over_token);
    }
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_wait_for_token_min_hold_time(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case DS485_TOKEN_MIN_HOLD_TIMEOUT_SIG:
        /* fall through */
    case DS485_TOKEN_MAX_HOLD_TIMEOUT_SIG: {
        if (me->findStation) {
            return Q_TRAN(&DS485_find_station);
        }
        if (me->restartBus) {
            return Q_TRAN(&DS485_startup_master);
        }
        return Q_TRAN(&DS485_hand_over_token);
    }
    case DS485_RX_TOKEN_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if ((DS485_GetDestination(rxMsgEvt) == me->deviceAddress) && (DS485_GetSource(rxMsgEvt) != me->deviceAddress)) {
            /* received token from another device - this is OK -> don't do nothing */
            Q_ASSERT(me->tokenMaxHoldTimeElapsed == FALSE);
            return Q_HANDLED();
        }
        /* fall through */
    }
    case DS485_RX_STARTUP_SIG:
        /* fall through */
    case DS485_RX_JOIN_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_DATA_SIG: {
        if (!me->messageSentAfterReceivingToken) {
            /* received a message which did not come from me -> restart */
            return Q_TRAN(&DS485_startup);
        }
        break;
    }
    case DS485_TX_MSG_SIG: {
        /* repost event */
        QActive_postLIFO((QActive *)me, e);
        return Q_TRAN(&DS485_transmit_data);
    }
    }
    return Q_SUPER(&DS485_holding_token);
}

/*..........................................................................*/
static QState DS485_no_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
#ifdef DAEMON_IMPL
        checkForMissingBusMembers(me);
#endif
        /* reset used addresses */
        memset(me->usedAddresses, 0, sizeof(me->usedAddresses));
        SetBitSet8(me->usedAddresses, me->deviceAddress);
        /* start token missing timeout */
        QTimeEvt_postIn(&me->tokenMissingTimeEvt, (QActive *)me, (me->master ? TOKEN_MISSING_MASTER_TIMEOUT_TICKS : TOKEN_MISSING_SLAVE_TIMEOUT_TICKS));
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->tokenMissingTimeEvt);
        return Q_HANDLED();
    case DS485_RX_TOKEN_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        /* update used addresses */
#ifdef DAEMON_IMPL
        if( !isAddressValid(me, DS485_GetSource(rxMsgEvt)) && IsBitSet8Cleared(me->dsidRequested, DS485_GetSource(rxMsgEvt)) ){
            DS485MessageEvt * txMsgEvt = DS485_CreateControlMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_GET_ADDRESS, 0);
            QActive_postFIFO(AO_DS485, (QEvent*)txMsgEvt);
            SetBitSet8(me->dsidRequested, DS485_GetSource(rxMsgEvt));
        }
#endif
        SetBitSet8(me->usedAddresses, DS485_GetSource(rxMsgEvt));
        /* check destination */
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* holding token */
            return Q_TRAN(&DS485_holding_token);
        }
        /* check source */
        if (DS485_GetSource(rxMsgEvt) == me->deviceAddress) {
            /* token with identical source address received */
            return Q_TRAN(&DS485_startup);
        }
        return Q_HANDLED();
    }
    case DS485_RX_GET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
                /* send get address response */
                DS485MessageEvt * txMsgEvt = DS485_CreateControlMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, DS485_TYPE_ACK_RES, DS485_CMD_GET_ADDRESS, sizeof(me->dsmid));
                memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
                DS485_SendMessage(txMsgEvt);
                ++me->statistics.txGetAddressResponse;
            }
        } else {
#ifdef DAEMON_IMPL
            if (DS485_GetSource(rxMsgEvt) != me->deviceAddress) {
                dsuid_t dsuid = ds485_dsid_to_dsuid((dsid_t*)&rxMsgEvt->message[DS485_DATA_OFFSET]);
                registerAddress(me, DS485_GetSource(rxMsgEvt), &dsuid);
                ClearBitSet8(me->dsidRequested, DS485_GetSource(rxMsgEvt));
            }
#endif
        }
        return Q_HANDLED();
    }
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
                /* save successor device address */
                me->successorDeviceAddress = rxMsgEvt->message[DS485_DATA_OFFSET];
                /* send set successor address response */
                DS485_SendSetSuccessorAddressResponse(me, DS485_GetSource(rxMsgEvt));
            }
        }
        return Q_HANDLED();
    }
    case DS485_RX_DATA_SIG: {
        DS485MessageEvt * rxMsgEvt = (DS485MessageEvt *)e;
#ifdef DAEMON_IMPL
        rxMsgEvt->sourceAddr = DS485_getDsuidByBusId(me, DS485_GetSource(rxMsgEvt));
#endif
        switch (DS485_GetCommandType(rxMsgEvt)) {
        case DS485_TYPE_REQ: {
            ++me->statistics.rxDataRequest;
            if ((!DS485_IsBroadcast(rxMsgEvt)) && (DS485_GetDestination(rxMsgEvt) == me->deviceAddress)) {
                /* send acknowledge */
                DS485_SendMessage(DS485_CreateControlMessage(DS485_GetSource(rxMsgEvt), me->deviceAddress, DS485_TYPE_ACK_RES, DS485_CMD_RESPONSE, 0));
                ++me->statistics.txResponse;
            }
            break;
        }
        case DS485_TYPE_ACK_RES: {
            ++me->statistics.rxDataResponse;
            break;
        }
        case DS485_TYPE_EVENT: {
            ++me->statistics.rxDataEvent;
            break;
        }
        default: {
            /* should never get here */
            break;
        }
        }
#ifdef DAEMON_IMPL
        if (!DS485_IsBroadcast(rxMsgEvt)) {
            rxMsgEvt->destinationAddr = DS485_getDsuidByBusId(me, DS485_GetDestination(rxMsgEvt));
        }
        /* post request to dSM API active object */
        QActive_postFIFO(AO_DsmApi, e);
#endif
        return Q_HANDLED();
    }
    case DS485_RX_RESPONSE_SIG: {
#ifdef DAEMON_IMPL
        DS485MessageEvt * rxMsgEvt = (DS485MessageEvt *)e;
        rxMsgEvt->sourceAddr = DS485_getDsuidByBusId(me, DS485_GetSource(rxMsgEvt));
        /* responses are not sent to broadcast */
        rxMsgEvt->destinationAddr = DS485_getDsuidByBusId(me, DS485_GetDestination(rxMsgEvt));
        /* post message to DSM API active object */
        QActive_postFIFO(AO_DsmApi, e);
#else
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* post message to DSM API active object */
            QActive_postFIFO(AO_DsmApi, e);
        }
#endif
        return Q_HANDLED();
    }
    case DS485_TOKEN_MISSING_TIMEOUT_SIG:
        if (me->master) {
            /* this device is master -> generate and hand over new token */
            if (me->restartBus) {
                return Q_TRAN(&DS485_startup_master);
            }
            return Q_TRAN(&DS485_hand_over_token);
        }
        /* this device is slave -> start from scratch */
        return Q_TRAN(&DS485_startup);
#ifdef DAEMON_IMPL
    case DS485_RX_JOIN_SIG: {
        /* check if current master is still known */
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsRequest(rxMsgEvt)) {
            dsuid_t dsuid = ds485_dsid_to_dsuid((dsid_t*)&rxMsgEvt->message[DS485_DATA_OFFSET]);
            if (cmpAddress(me, 0, &dsuid) != 0) {
                registerAddress(me, 0, &dsuid);
            }
        }
        break;
    }
#endif
    case DS485_RX_STARTUP_SIG: {
        /* bus might be in restart mode, accept STARTUP to restart missing token timer */
        if (!me->master) {
            QTimeEvt_rearm(&me->tokenMissingTimeEvt, TOKEN_MISSING_SLAVE_TIMEOUT_TICKS);
            return Q_HANDLED();
        }
        break;
    }
    }
    return Q_SUPER(&DS485_operational);
}

/*..........................................................................*/
static QState DS485_hand_over_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        me->tokenHandoverCount = 0;
        return Q_HANDLED();
    case Q_INIT_SIG:
        return Q_TRAN(&DS485_send_token);
    }
    return Q_SUPER(&DS485_operational);
}

/*..........................................................................*/
static QState DS485_send_token(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        ++me->tokenHandoverCount;
        /* send token to successor device */
        DS485_SendToken(me->successorDeviceAddress, me->deviceAddress);
        ++me->statistics.txToken;
        QTimeEvt_postIn(&me->handOverTokenTimeEvt, (QActive *)me, TOKEN_HAND_OVER_TIMEOUT_TICKS);
        return Q_HANDLED();
    case Q_EXIT_SIG:
        QTimeEvt_disarm(&me->handOverTokenTimeEvt);
        return Q_HANDLED();
    case DS485_HAND_OVER_TOKEN_TIMEOUT_SIG:
        if (me->tokenHandoverCount == 3) {
            /* previous requested successor was not available */
            ClearBitSet8(me->usedAddresses, me->successorDeviceAddress);
            /* reset hand over count */
            me->tokenHandoverCount = 0;
            /* update statistics */
            ++me->statistics.successorLost;
            if (me->successorDeviceAddress != me->deviceAddress) {
                /* successor device lost */
                QActive_postFIFO(AO_MMI, &l_ds485DeviceLostEvt);
                /* check if unavailable successor is master */
                if (me->successorDeviceAddress == DS485_MASTER_ADDRESS) {
                    /* master did not respond -> I am master now */
                    DS485_ChangeToMaster(me);
                }
                /* get next successor station */
                me->successorDeviceAddress = DS485_GetNextSuccessorAddress(me, me->successorDeviceAddress);
                /* find successor device */
                return Q_TRAN(&DS485_find_successor_device);
            }
        }
        /* re-send token */
        return Q_TRAN(&DS485_send_token);
    case DS485_RX_TOKEN_SIG: {
        /* check destination address */
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_GetDestination(rxMsgEvt) == me->deviceAddress) {
            /* the token is for me */
            if (!me->terminate) {
                /* repost message */
                QActive_postLIFO((QActive *)me, e);
                /* change to no token state */
                return Q_TRAN(&DS485_no_token);
            }
            /* terminate */
            return Q_TRAN(&DS485_stopped);
        }
        /* fall through */
    }
    case DS485_RX_STARTUP_SIG:
        /* fall through */
    case DS485_RX_JOIN_SIG:
        /* fall through */
    case DS485_RX_GET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_SET_SUCCESSOR_ADDRESS_SIG:
        /* fall through */
    case DS485_RX_RESPONSE_SIG:
        /* fall through */
    case DS485_RX_DATA_SIG: {
        /* check source address */
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_GetSource(rxMsgEvt) != me->deviceAddress) {
            /* somebody else is active */
            if (!me->terminate) {
                /* repost message */
                QActive_postLIFO((QActive *)me, e);
                /* change to no token state */
                return Q_TRAN(&DS485_no_token);
            }
            /* terminate */
            return Q_TRAN(&DS485_stopped);
        }
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&DS485_hand_over_token);
}

/*..........................................................................*/
static QState DS485_find_successor_device(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG:
        /* send get address request (ping) */
        DS485_SendMessage(DS485_CreateControlMessage(me->successorDeviceAddress, me->deviceAddress, DS485_TYPE_REQ, DS485_CMD_GET_ADDRESS, 0));
        ++me->statistics.txGetAddressRequest;
        /* start timer */
        QTimeEvt_postIn(&me->handOverTokenTimeEvt, (QActive *)me, GET_ADDRESS_REQUEST_TIMEOUT_TICKS);
        return Q_HANDLED();
    case Q_EXIT_SIG:
        /* stop timer */
        QTimeEvt_disarm(&me->handOverTokenTimeEvt);
        return Q_HANDLED();
    case DS485_RX_GET_ADDRESS_SIG: {
        DS485MessageEvt const * rxMsgEvt = (DS485MessageEvt const *)e;
        if (DS485_IsResponse(rxMsgEvt)) {
            /* get address response received */
            if (DS485_GetSource((DS485MessageEvt const *)e) == me->successorDeviceAddress) {
                /* we got a ping response -> send token to this new successor */
                return Q_TRAN(&DS485_send_token);
            }
        }
        return Q_HANDLED();
    }
    case DS485_HAND_OVER_TOKEN_TIMEOUT_SIG:
        /* previous requested successor was not available */
        ClearBitSet8(me->usedAddresses, me->successorDeviceAddress);
        /* check if unavailable successor is master */
        if (me->successorDeviceAddress == DS485_MASTER_ADDRESS) {
            /* master did not respond -> I am master now */
            DS485_ChangeToMaster(me);
            return Q_TRAN(&DS485_send_token);
        }
        /* find next station */
        me->successorDeviceAddress = DS485_GetNextSuccessorAddress(me, me->deviceAddress);
        if (me->deviceAddress == me->successorDeviceAddress) {
#ifdef DAEMON_IMPL
            /* clean address lookup table */
            checkForMissingBusMembers(me);
#endif
            return Q_TRAN(&DS485_startup_master);
        } else {
            return Q_TRAN(&DS485_find_successor_device);
        }
    }
    return Q_SUPER(&DS485_hand_over_token);
}

/*..........................................................................*/
static QState DS485_stopped(DS485 *me, QEvent const *e)
{
    switch (e->sig) {
    case Q_ENTRY_SIG: {
        /* send DS485 stopped event */
        static const QEvent ds485StoppedEvt = { DS485_STOPPED_SIG, 0 };
        QActive_postFIFO(AO_MMI, &ds485StoppedEvt);
        return Q_HANDLED();
    }
    }
    return Q_SUPER(&QHsm_top);
}

/*..........................................................................*/
static void DS485_SendToken(uint8_t dest, uint8_t src)
{
    DS485MessageEvt* msgEvt = Q_NEW(DS485MessageEvt, DS485_TX_MSG_SIG);
    uint8_t* msg = msgEvt->message;
    /* build header */
    msg[0] = DS485_START_BYTE;
    msg[1] = dest;
    msg[2] = src;
    msg[3] = (DS485_CMD_TOKEN << 4);
    /* set message size */
    msgEvt->size = DS485_HEADER_CRC_SIZE;

    DS485_SendMessage( msgEvt );
}

/*..........................................................................*/
static void DS485_SendMessage(DS485MessageEvt *msgEvt)
{
    uint8_t * message = msgEvt->message;
    /* calculate crc16 ccitt */
    uint16_t crc = crc16_ccitt_calc_data(0, message, msgEvt->size-2);
    /* add crc16 (most significant byte first) */
    message[msgEvt->size-1] = (uint8_t)(crc & 0xFF);
    message[msgEvt->size-2] = (uint8_t)(crc >> 8);
    /* post message to queue */
    QEQueue_postFIFO(&l_ds485TxMessageQueue, (QEvent const *)msgEvt);
    /* enable transmit interrupt */
#ifdef DAEMON_IMPL
    ds485_system_tx();
#else
    USART3_EnableTransmitInterrupt();
#endif
}

/*..........................................................................*/
static void DS485_SendSetAddressResponse(DS485 *me, uint8_t dest)
{
    /* send set successor address response */
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DS485_CreateControlMessage(dest, me->deviceAddress, DS485_TYPE_ACK_RES, DS485_CMD_SET_ADDRESS, sizeof(me->dsmid));
    memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
    DS485_SendMessage(txMsgEvt);
    ++me->statistics.txSetAddressResponse;
}

/*..........................................................................*/
static void DS485_SendSetSuccessorAddressResponse(DS485 *me, uint8_t dest)
{
    /* send set successor address response */
    DS485MessageEvt * txMsgEvt;
    txMsgEvt = DS485_CreateControlMessage(dest, me->deviceAddress, DS485_TYPE_ACK_RES, DS485_CMD_SET_SUCCESSOR_ADDRESS, sizeof(me->dsmid)+1);
    memcpy(&txMsgEvt->message[DS485_DATA_OFFSET], &me->dsmid, sizeof(me->dsmid));
    txMsgEvt->message[DS485_DATA_OFFSET+sizeof(me->dsmid)] = me->successorDeviceAddress;
    DS485_SendMessage(txMsgEvt);
    ++me->statistics.txSetSuccessorAddressResponse;
}

/*..........................................................................*/
static void DS485_PublishDeviceAddress(uint8_t deviceAddress)
{
    DS485AddressEvt * addrEvt = Q_NEW(DS485AddressEvt, DS485_ADDRESS_SIG);
    addrEvt->deviceAddress = deviceAddress;
    QF_publish((QEvent *)addrEvt);
}

/*..........................................................................*/
static void DS485_ChangeToMaster(DS485 *me)
{
    /* disarm find station time event */
    QTimeEvt_disarm(&me->findStationTimeEvt);
    /* set master flag */
    me->master = TRUE;
    /* set find station flag */
    me->findStation = TRUE;
    /* change own device address */
    ClearBitSet8(me->usedAddresses, me->deviceAddress);
    me->deviceAddress = DS485_MASTER_ADDRESS;
    SetBitSet8(me->usedAddresses, me->deviceAddress);
#ifdef DAEMON_IMPL
    dsuid_t dsuid = ds485_dsid_to_dsuid(&me->dsmid);
    registerAddress(me, me->deviceAddress, &dsuid);
#endif
    /* publish device address */
    DS485_PublishDeviceAddress(me->deviceAddress);
}

/*..........................................................................*/
static uint8_t DS485_GetNextSuccessorAddress(DS485 *me, uint8_t startDeviceAddress)
{
    uint8_t deviceAddress = startDeviceAddress;
    do {
        /* find next successor station */
        ++deviceAddress;
    } while (IsBitSet8Cleared(me->usedAddresses, deviceAddress) && (deviceAddress <= DS485_MAX_SLAVE_ADDRESS));
    if (deviceAddress > DS485_MAX_SLAVE_ADDRESS) {
        deviceAddress = DS485_MASTER_ADDRESS;
    }
    return deviceAddress;
}

/*..........................................................................*/
static void DS485_ResetStatistics(DS485 *me)
{
    memset(&me->statistics, 0, sizeof(me->statistics));
}

/*..........................................................................*/
static void DS485_RequestStatistics(DS485 *me, DS485ReqStatisticsEvt const * reqEvt)
{
    DS485ResStatisticsEvt * resEvt = Q_NEW(DS485ResStatisticsEvt, DS485_RES_STATISTICS_SIG);
    memcpy(&resEvt->statistics, &me->statistics, sizeof(resEvt->statistics));
    QActive_postFIFO(reqEvt->requester, (QEvent *)resEvt);
}

#ifdef DAEMON_IMPL
static uint8_t DS485_getBusIdByDsuid( DS485 *me, dsuid_t *dsuid )
{
    int i;
    for( i = 0; i < sizeof(me->busDsuids)/sizeof(me->busDsuids[0]); i++ ){
        if( cmpAddress(me, i, dsuid) == 0 ){
            return  i;
        }
    }
    return  DS485_INVALID_ADDRESS;
}

static dsuid_t DS485_getDsuidByBusId( DS485 *me, uint8_t busId )
{
    return me->busDsuids[busId];
}
#endif
